/************************************************************************/
/*									*/
/*  Procedures to read an ispell affix file and dictionary files and	*/
/*  to build a finite automaton that recognises the same words.		*/
/*									*/
/*  Treatment of ispell affix files.					*/
/*									*/
/*  This source code is not meant as a replacement for the ispell	*/
/*  material, it is just to make things fit in with what I want to do	*/
/*  with it better.							*/
/*									*/
/*  Syntax is simplified. (No line breaks allowed in rule.)		*/
/*									*/
/************************************************************************/

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<string.h>
#   include	<ctype.h>
#   include	<charnames.h>
#   include	<ind.h>
#   include	<uniUtf8.h>
#   include	<ucd.h>
#   include	<appDebugon.h>
#   include	"affix.h"

/************************************************************************/
/*									*/
/*  Debugging routines.							*/
/*									*/
/************************************************************************/

static void indAffixListString(	const RuleString *	rs )
    {
    int		i;

    if  ( rs->rsLength == 0 )
	{ appDebug( "%c", '-' ); return;	}

    for ( i= 0; i < rs->rsLength; i++ )
	{
	unsigned char	scratch[10];
	int		step;
	int		j;

	step= uniPutUtf8( scratch, rs->rsUtf16[i] );
	if  ( step < 1 )
	    { LLLDEB(i,rs->rsUtf16[i],step);	}

	for ( j= 0; j < step; j++ )
	    { appDebug( "%c", scratch[j] );	}
	}

    return;
    }

static int indAffixListMaskChar(	int		val,
					void *		through )
    {
    unsigned char	scratch[10];
    int			step;
    int			j;

    step= uniPutUtf8( scratch, val );
    if  ( step < 1 )
	{ LLDEB(val,step);	}

    for ( j= 0; j < step; j++ )
	{ appDebug( "%c", scratch[j] );	}

    return 0;
    }

void indAffixListMask(	const ApplicabilityMask *	am )
    {
    int		n;
    IndexSet	is;

    utilInitIndexSet( &is );
    n= utilIndexSetIntersect( &is, &(am->amSet), &(am->amSet) );
    utilCleanIndexSet( &is );

    if  ( n < 0 )
	{ LDEB(n); return;	}

    if  ( n == 0 )
	{
	if  ( am->amNegative )
	    { appDebug( "%c", '.' );	}
	else{ appDebug( "[^.]" );	}

	return;
	}

    if  ( am->amNegative || n > 1 )
	{ appDebug( "%c", '[' );	}
    if  ( am->amNegative )
	{ appDebug( "%c", '^' );	}
    utilIndexSetForAll( &(am->amSet), indAffixListMaskChar, (void *)0 );
    if  ( am->amNegative || n > 1 )
	{ appDebug( "%c", ']' );	}

    return;
    }

static void indAffixListMasks(	const AffixRule *	ar )
    {
    int		i;

    for ( i= 0; i < ar->arMaskLength; i++ )
	{ indAffixListMask( ar->arMasks+ i );	}

    return;
    }

void indAffixListRule(	int			rule,
			const AffixRule *	ar )
    {
    if  ( ar->arWhatSide > 0 )
	{ appDebug( "suffix flag " );	}
    else{ appDebug( "prefix flag " );	}

    if  ( ar->arCross )
	{ appDebug( "*" );	}

    if  ( rule >= 0 )
	{
	if  ( isprint( rule ) )
	    { appDebug( "%c: ", rule );	}
	else{ appDebug( "%d: ", rule );	}
	}

    indAffixListMasks( ar );

    appDebug( " > " );

    if  ( ar->arReplaceFrom.rsLength > 0 )
	{
	appDebug( "-" );
	indAffixListString( &(ar->arReplaceFrom) );
	appDebug( "," );
	}

    indAffixListString( &(ar->arReplaceWith) );
    appDebug( "\n" );

    return;
    }

void indAffixListRules(	const void * voidars )
    {
    const AffixRules *		ars= (AffixRules *)voidars;
    const AffixRule *		ar;
    int				rule;

    appDebug( "AFFIX RULES:\n" );

    for ( rule= 0; rule < 256; rule++ )
	{
	ar= ars->arsRules[rule];

	while( ar )
	    { indAffixListRule( rule, ar ); ar= ar->arNext;	}
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Catinate two strings while applying rules.				*/
/*									*/
/************************************************************************/

static void affixCatinate(	unsigned short *	scratch,
				const unsigned short *	w1,
				int			l1,
				const unsigned short *	w2,
				int			l2 )
    {
    int		i;

    for ( i= 0; i < l1; i++ )
	{ *(scratch++)= *(w1++);	}
    for ( i= 0; i < l2; i++ )
	{ *(scratch++)= *(w2++);	}
    *scratch= 0;

    return;
    }

static int affixMatch(		const unsigned short *	w,
				const AffixRule *	ar )
    {
    int				j;
    const ApplicabilityMask *	am= ar->arMasks;

    am= ar->arMasks;
    for ( j= 0; j < ar->arMaskLength; w++, am++, j++ )
	{
	if  ( am->amNegative )
	    {
	    if  (   utilIndexSetContains( &(am->amSet), *w ) )
		{ return 0;	}
	    }
	else{
	    if  ( ! utilIndexSetContains( &(am->amSet), *w ) )
		{ return 0;	}
	    }
	}

    return 1;
    }

static int affixApplySuffixRule(	const AffixRules *	ars,
					unsigned short *	scratch,
					const unsigned short *	word,
					int			len,
					const AffixRule *	ar,
					void *			ind )
    {
    if  ( ! affixMatch( word+ ( len- ar->arMaskLength ), ar ) )
	{ return 0;	}

    if  ( ar->arSuffixNumber >= 0 )
	{
	affixCatinate( scratch,
		    word, len- ar->arReplaceFrom.rsLength,
		    ar->arReplaceWith.rsUtf16, 0 );

	if  ( indAddSuffixUtf16( ind, scratch, ar->arSuffixNumber ) )
	    { LDEB(ar->arSuffixNumber); return -1;	}

	return 0;
	}

    affixCatinate( scratch,
		    word, len- ar->arReplaceFrom.rsLength,
		    ar->arReplaceWith.rsUtf16, ar->arReplaceWith.rsLength );

    if  ( ar->arShift > 0 && ucdIsLo( scratch[0] ) )
	{ scratch[0]= ucdToUpper( scratch[0] ); }
    if  ( ar->arShift < 0 && ucdIsLu( scratch[0] ) )
	{ scratch[0]= ucdToLower( scratch[0] ); }

    if  ( indPutUtf16( ind, scratch ) < 0 )
	{ XDEB(scratch); return -1;	}

    return 1;
    }

static int affixApplyPrefixRule(	const AffixRules *	ars,
					unsigned short *	scratch,
					const unsigned short *	word,
					int			len,
					const AffixRule *	ar,
					void *			ind )
    {
    if  ( ! affixMatch( word, ar ) )
	{ return 0;	}

    affixCatinate( scratch,
	ar->arReplaceWith.rsUtf16, ar->arReplaceWith.rsLength,
	word+ ar->arReplaceFrom.rsLength, len- ar->arReplaceFrom.rsLength );

    if  ( ar->arShift > 0 && ucdIsLo( scratch[0] ) )
	{ scratch[0]= ucdToUpper( scratch[0] ); }
    if  ( ar->arShift < 0 && ucdIsLu( scratch[0] ) )
	{ scratch[0]= ucdToLower( scratch[0] ); }

    if  ( indPutUtf16( ind, scratch ) < 0 )
	{ XDEB(scratch); return -1;	}

    return 1;
    }

static int affixApplySuffixRules(	const AffixRules *	ars,
					int			all,
					const unsigned short *	word,
					int			len,
					const char *		rules,
					void *			ind )
    {
    AffixRule *		ar;
    int			i;

    unsigned short	scratch[200+1];

    i= 0;
    while( rules[i] )
	{
	ar= ars->arsRules[rules[i]&0xff];

	while( ar )
	    {
	    if  ( ( all || ar->arCross )	&&
		  ar->arWhatSide > 0		&&
		  ar->arMaskLength <= len	)
		{
		if  ( affixApplySuffixRule( ars, scratch, word, len,
							    ar, ind ) < 0 )
		    { XDEB(word); return -1;	}
		}

	    ar= ar->arNext;
	    }

	i++;
	}

    return 0;
    }

int affixApplyPrefixRules(		const AffixRules *	ars,
					int			cross,
					const unsigned short *	word,
					int			len,
					const char *		rules,
					void *			ind )
    {
    AffixRule *		ar;
    int			i;

    unsigned short	scratch[200+1];

    i= 0;
    while( rules[i] )
	{
	ar= ars->arsRules[rules[i]&0xff];

	while( ar )
	    {
	    if  ( ar->arWhatSide < 0		&&
		  ar->arMaskLength <= len	)
		{
		int done= affixApplyPrefixRule( ars, scratch, word, len,
								    ar, ind );
		if  ( done < 0 )
		    { SDEB((char *)word); return -1;	}

		if  ( cross && done == 1 && ar->arCross )
		    {
		    const int	all= 0;

		    if  ( affixApplySuffixRules( ars, all, word, len,
							    rules, ind ) < 0 )
			{ LDEB(done); return -1;	}
		    }
		}

	    ar= ar->arNext;
	    }

	i++;
	}

    return 0;
    }

int indAffixApplyRules(	const AffixRules *	ars,
			int			prefixOnly,
			const unsigned short *	word,
			int			len,
			const char *		rules,
			void *			ind )
    {
    if  ( affixApplyPrefixRules( ars, ! prefixOnly, word, len, rules, ind ) )
	{ LDEB(1); return -1;	}

    if  ( ! prefixOnly )
	{
	const int	all= 1;

	if  ( affixApplySuffixRules( ars, all, word, len, rules, ind ) )
	    { LDEB(1); return -1;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Build suffix administration.					*/
/*									*/
/************************************************************************/

int affixCollectSuffixes(	AffixThrough *		at )
    {
    int		r;

    for ( r= 0; r < 256; r++ )
	{
	AffixRule *		ar= at->atRules->arsRules[r];

	while( ar )
	    {
	    RuleString		rs;
	    int			f;
	    int			t;

	    if  ( ar->arWhatSide > 0 )
		{
		/*  swap */
		rs= ar->arReplaceWith;
		t= rs.rsLength- 1;
		for ( f= 0; f < rs.rsLength/ 2; f++, t-- )
		    {
		    int		s;

		    s= rs.rsUtf16[f];
		    rs.rsUtf16[f]= rs.rsUtf16[t];
		    rs.rsUtf16[t]= s;
		    }

		f= indPutUtf16( at->atSuffixInd, rs.rsUtf16 );
		if  ( f < 0 )
		    { LDEB(f); return -1;	}

		t= utilIndexMappingGet( &(at->atSuffixMapping), f );
		if  ( t < 0 )
		    {
		    t= indPutSuffixUtf16( at->atVoidInd,
						    ar->arReplaceWith.rsUtf16 );
		    if  ( t < 0 )
			{ LDEB(t); return -1;	}

		    if  ( utilIndexMappingPut( &(at->atSuffixMapping), f, t ) )
			{ LLDEB(f,t); return -1;	}
		    }
		ar->arSuffixNumber= t;
		}

	    ar= ar->arNext;
	    }
	}

    return 0;
    }

